
LOADTRACKS = "/loadtracks?identifier={query}"
DECODETRACK = "/decodetrack?track={track}"
DECODETRACKS = "/decodetracks"
ROUTEPLANNER = "/routeplanner/status"
UNMARK_FAILED_ADDRESS = "/routeplanner/free/address"
UNMARK_ALL_FAILED_ADDRESS = "/routeplanner/free/all"
